-- Copyright (c) 2012 Cloudera, Inc. All rights reserved.
--
-- Version 43 of the CMF schema. Migration code to be applied on top of the
-- previous version.
--

update SCHEMA_VERSION set VERSION=43;

-- Modifications for a config container and for host configs container

--
-- Create a new table to store generic config containers
--
create table CONFIG_CONTAINERS (
    CONFIG_CONTAINER_ID number constraint NOT_NULL_CONFIG_CONTAINER_ID not null,
    OPTIMISTIC_LOCK_VERSION number constraint NOT_NULL_CFG_CTR_OPT_LOCK_VER not null,
    CONFIG_TYPE varchar2(255) constraint NOT_NULL_CFG_CTR_CONFIG_TYPE not null constraint UNIQUE_CFG_CTR_CONFIG_TYPE unique,
    CONFIG_REVISION_ID number,
    constraint PK_CONFIG_CONTAINERS primary key (CONFIG_CONTAINER_ID)
);

--
-- Create a new column in CONFIG_REVISIONS to store config container id
--
alter table CONFIG_REVISIONS
      add CONFIG_CONTAINER_ID number;

--
-- Create a new column in HOSTS to store config container id
--
alter table HOSTS
      add CONFIG_CONTAINER_ID number;

--
-- Create a new column in CONFIGS to store config container id
--
alter table CONFIGS
      add CONFIG_CONTAINER_ID number;

--
-- Add a column in the AUDITS table to refer to a config container to
-- facilitate auditing of updates to configs 'own'-ed by the config container.
--
alter table AUDITS
      add CONFIG_CONTAINER_ID number;

--
-- Drop obsolete foreign key constraints
--
alter table CONFIG_REVISIONS
      drop constraint FK_CONFIG_REVISION_HOST;

alter table HOSTS
      drop constraint FK_HOST_CONFIG_REVISION;

--
-- Clean up unwanted columns from the config revision table.
-- Remove existing config revisions that refer to hosts since we no
--   longer have config revisions associated with hosts.
-- Drop the host id column
--
delete from CONFIG_REVISIONS where HOST_ID is not null;
alter table CONFIG_REVISIONS
      drop column HOST_ID;

--
-- Clean up unwanted columns from the hosts table.
-- Drop the config revision id column
--
alter table HOSTS
      drop column CONFIG_REVISION_ID;

--
-- Insert a config container to track host configs.
--
insert into CONFIG_CONTAINERS (CONFIG_CONTAINER_ID, OPTIMISTIC_LOCK_VERSION, CONFIG_TYPE)
       values (1, 1, 'ALL_HOSTS');

-- Insert a base revision for existing config containers.
-- Makes sure that the REVISION_IDs used for config containers are greater than
-- any existing REVISION_IDs in the CONFIG_REVISIONS table.
-- The value used for the CREATED_INSTANT is the UTC based epoch time, it
-- is equivalent to MySql 'unix_timestamp()'
-- 86400000 = 24 * 60 * 60 * 1000, number of milliseconds in a day
insert into CONFIG_REVISIONS (REVISION_ID, CONFIG_CONTAINER_ID, CREATED_INSTANT, MESSAGE, VISIBLE)
      select LAST_REV.MAX_ID + rownum REVISION_ID,
             cc.CONFIG_CONTAINER_ID,
             (cast(sys_extract_utc(systimestamp) as date) - date'1970-01-01') * 86400000 CREATED_INSTANT,
             'Base revision' MESSAGE,
             1
      from CONFIG_CONTAINERS cc,
           (select nvl(max(REVISION_ID), 0) MAX_ID from CONFIG_REVISIONS) LAST_REV;

--
-- Link each config container to its current config revision.
--
update CONFIG_CONTAINERS cc
      set CONFIG_REVISION_ID=
            (select REVISION_ID
	    from CONFIG_REVISIONS r
	    where cc.CONFIG_CONTAINER_ID=r.CONFIG_CONTAINER_ID);

--
-- Link each host to the hosts config container
--
update HOSTS h
      set CONFIG_CONTAINER_ID=
      (select CONFIG_CONTAINER_ID
      from CONFIG_CONTAINERS cc
      where cc.CONFIG_TYPE='ALL_HOSTS');

--
-- All data insertion is done. Set up, update indices and constraints.
-- Note that CONFIG_CONTAINER and CONFIG_REVISION have been shortened
-- to CFG_CONTAINER and CFG_REVISION in the index the foreign key
-- constraint names to avoid running into the limit for identifier names.
--
create index IDX_CONFIG_CFG_CONTAINER on CONFIGS (CONFIG_CONTAINER_ID);

alter table CONFIGS
      add constraint FK_CONFIG_CFG_CONTAINER
      foreign key (CONFIG_CONTAINER_ID)
      references CONFIG_CONTAINERS
      deferrable initially immediate;

create index IDX_CFG_REVISION_CFG_CONTAINER on CONFIG_REVISIONS (CONFIG_CONTAINER_ID);

alter table CONFIG_REVISIONS
      add constraint FK_CFG_REVISION_CFG_CONTAINER
      foreign key (CONFIG_CONTAINER_ID)
      references CONFIG_CONTAINERS
      deferrable initially immediate;

create index IDX_CFG_CONTAINER_CFG_REVISION on CONFIG_CONTAINERS (CONFIG_REVISION_ID);

alter table CONFIG_CONTAINERS
      add constraint FK_CFG_CONTAINER_CFG_REVISION
      foreign key (CONFIG_REVISION_ID)
      references CONFIG_REVISIONS
      deferrable initially immediate;

create index IDX_HOST_CFG_CONTAINER on HOSTS (CONFIG_CONTAINER_ID);

alter table HOSTS
      add constraint FK_HOST_CFG_CONTAINER
      foreign key (CONFIG_CONTAINER_ID)
      references CONFIG_CONTAINERS
      deferrable initially immediate;
