// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
	"cloudera/chart/AnyStockLineChartSet",
	"cloudera/chart/TimeRange",
	"cloudera/chart/ChartStretcher",
	"cloudera/Util",
	"cloudera/chart/lang/AllLocales"
], function(AnyStockLineChartSet, TimeRange, ChartStretcher, Util, AllLocales) {

	/**
	 * The base class for dealing with charts using contexts and metrics.
	 *
	 * This design requires a fixed number of dataSet slots, and
	 * a fixed number of dataProvider slots.
	 *
	 * Whenever we have filled all the slots, and need more room, we expand
	 * the number of dataSet slots. Note: this also need to trigger a full
	 * refresh on the AnyChart's side.
	 *
	 * The number of dataSet slots should equal to the number of dataProvider slots.
	 */
	return AnyStockLineChartSet.extend({

		_preinitialize: function() {
			this._super.apply(this, arguments);
			this.setShowEventMarkers(true);
			this.allowPadding = false;
			this.allowChartsToStretch = true;
		},

		createScroller: function() {
			if (this.chartStretcher) {
				return this.chartStretcher.createScroller();
			} else {
				return {
					enabled: false
				};
			}
		},

		createSettings: function(){
			var localeSettings = AllLocales.getLocale(Util.getLocale()).settings;
			var settings = {
				'dataGrouping':{
					'enabled':false
				},
				'tooltipSettings': {
					'enabled': true
				},
				'charts': this.createCharts(),
				'scroller': this.createScroller(),
				'valueAxes': {
					'primary': {
						'enabled': true,
						'scale': {
							'minimumMode': 'CustomValue',
							'minimum': 0
						}
					}
				},
				'timeScale': {
					'isOrdinal':false,
					'selectedRange': {
						'type':'Max'
					},
					'majorIntervalFormats': localeSettings.timeScale.majorIntervalFormats,
					'minorIntervalFormats': localeSettings.timeScale.minorIntervalFormats
				},
				'displayLocale': localeSettings.displayLocale
			};
			return settings;
		},

		createCharts: function() {
			return [];
		},

		/**
		 * Creates a chart.
		 * @param chartId - identifies the chart.
		 * @param chartTitle - the chart title (appears on the top left corner).
		 * @param units - see metric.json for details
		 * @param derivative - string of whether this is a derivative or not
		 * @see AnyStock for details of the format.
		 */
		createChart: function(chartId, chartTitle, units, derivative) {
			var localeSettings = AllLocales.getLocale(Util.getLocale()).settings;
			var chart = {
				'enabled':true,
				'id': chartId,
				'legend': {
					'title': {
						'enabled': true,
						'format': chartTitle,
						'focusSettings': {
							'mouseOver': {
								'format': chartTitle
							},
							'mouseOut': {
								'format': chartTitle
							}
						},
						'font': {
							'color': '#262626',
							'size': 12
						}
					},
					'dateTime': {
						'enabled': false
					},
					'background':{
						'enabled':true,
						'opacity':1,
						'fill':{
							'color':'#FFFFFF',
							'enabled':true,
							'type':'Solid'
						}
					}
				},
				'valueAxes':{
					'primary': {
						'position' :"Left",
						'labels': {
							'position': "Inside",
							'padding': 3,
							'valign': "Center",
						        'format': this.formatUnits(units, derivative)
						},
						'scale': {
							'minimumMode': 'CustomValue',
							'minimum': 0
						},
						'line': {
							'enabled':false
						},
						'grid': {
							'enabled':false
						},
						'zeroLine': {
							'enabled':false
						}
					},
					'extras': []
				},
				'seriesList': this.createSeriesList(),
				"xAxis": {
					"dateHighlighter": {
						"line": {
							"thickness": 1,
							"color": "#777777",
							"dashed": false,
							"opacity": 1
						},
						"labels": [ {
							"anchor": "Bottom",
							"xPadding": 0,
							"yPadding": 0,
							"valign": "Bottom",
							"halign": "Center",
							"format": "{%Date}{\"" + localeSettings.timeScale.intervalFormats.seconds + "\"}",
							"font": {
								"family": "Tahoma",
								"size": 9,
								"bold": true,
								"color": "#27413E"
							},
							"background": {
								"enabled": true,
								"fill": {
									"enabled": true,
									"type": "Gradient",
									"opacity": 1,
									"gradient": {
										"angle": 90,
										"keys": [ {
											"color": "#EDFFFF"
										}, {
											"color": "#A9E5FF"
										} ]
									}
								},
								"border": {
									"enabled": true,
									"color": "#0080E2",
									"opacity": 0.3
								},
								"insideMargin": {
									"left": 3,
									"right": 3,
									"bottom": -1
								}
							}
						} ]
					}
				}
			};
			return chart;
		},

		/**
		 * Converts the metric's units attribute into a format string
		 * that AnyStockChart understands.
		 *
		 * @param units - see metric.json for details
		 * @param derivative - string of whether this is a derivative or not
		 */
	        formatUnits: function(units, derivative) {
			var result = "{%Value}";
			var decimals = 0;
			var scale = "";
			if (units === "agents") {
				decimals = 0;
			} else if (units === "batches") {
				decimals = 0;
			} else if (units === "bytes") {
				decimals = 1;
				scale += ",scale:(1)(1024)(1024)(1024)(1024)(1024)|(B)(KiB)(MiB)(GiB)(TiB)(PiB)";
			} else if (units === "CPUs /s") {
				decimals = 1;
			} else if (units === "events") {
				decimals = 0;
			} else if (units === "fds") {
				decimals = 0;
			} else if (units === "threads") {
				decimals = 0;
			} else if (units === "GB") {
				decimals = 1;
				scale += ",scale:(1)(1024)(1024)|(GiB)(TiB)(PiB)";
			} else if (units === "groups") {
				decimals = 0;
			} else if (units === "items") {
				decimals = 0;
			} else if (units === "jobs") {
				decimals = 0;
			} else if (units === "connections") {
				decimals = 0;
			} else if (units === "maps") {
				decimals = 0;
			} else if (units === "megabyte") {
				decimals = 1;
				scale = ",scale:(1)(1024)(1024)(1024)|(MiB)(GiB)(TiB)(PiB)";
			} else if (units === "megabytes") {
				decimals = 1;
				scale = ",scale:(1)(1024)(1024)(1024)|(MiB)(GiB)(TiB)(PiB)";
			} else if (units === "ms") {
				decimals = 1;
				scale = ",scale:(1)(1000)(60)(60)(24)|(ms)(s)(min)( hour)( day)";
			} else if (units === "micros") {
				decimals = 1;
				scale = ",scale:(1)(1000)(1000)(60)(60)(24)|(micros)(ms)(s)(min)( hour)( day)";
			} else if (units === "DataNodes") {
				decimals = 0;
			} else if (units === "NameNodes") {
				decimals = 0;
			} else if (units === "pages") {
				decimals = 0;
			} else if (units === "network interfaces") {
				decimals = 0;
			} else if (units === "blocks") {
				decimals = 1;
				scale = ",scale:(1)(1000)(1000)|( blocks)(k blocks)(m blocks)";
			} else if (units === "percent") {
				decimals = 1;
				scale = "%";
			} else if (units === "records") {
				decimals = 0;
			} else if (units === "Records") {
				decimals = 0;
			} else if (units === "reduces") {
				decimals = 0;
			} else if (units === "seconds") {
				decimals = 1;
				scale = ",scale:(1)(60)(60)(24)|(s)(min)( hour)( day)";
			} else if (units === "nodes") {
				decimals = 0;
			} else if (units === "servers") {
				decimals = 0;
			} else if (units === "queries") {
				decimals = 0;
			} else if (units === "nodes") {
				decimals = 0;
			} else if (units === "assignments") {
				decimals = 0;
			} else if (units === "journalnodes") {
				decimals = 0;
			} else if (units === "volumes") {
				decimals = 0;
			} else if (units === "backends") {
				decimals = 0;
			} else if (units === "directories") {
				decimals = 0;
			} else if (units === "events") {
				decimals = 0;
			} else if (units === "messages") {
				decimals = 0;
			} else if (units === "Masters") {
				decimals = 0;
			} else if (units === "epoch") {
				decimals = 0;
			} else if (units === "transaction id") {
				decimals = 0;
			} else if (units === "attempts") {
				decimals = 0;
			} else if (units === "updates") {
				decimals = 0;
			} else if (units === "sinks") {
				decimals = 0;
			} else if (units === "sources") {
				decimals = 0;
			} else if (units === "transactions") {
				decimals = 0;
			} else if (units === "transceivers") {
				decimals = 0;
			} else if (units === "heartbeats") {
				decimals = 0;
			} else if (units === "slots") {
				decimals = 0;
			} else if (units === "replicas") {
				decimals = 0;
			} else if (units === "operations") {
				decimals = 0;
			} else if (units === "requests") {
				decimals = 0;
			} else if (units === "packets") {
				decimals = 0;
			} else if (units === "Impala Daemons") {
				decimals = 0;
			} else if (units === "errors") {
				decimals = 0;
			} else if (units === "collisions") {
				decimals = 0;
			} else if (units === "reads") {
				decimals = 0;
			} else if (units === "writes") {
				decimals = 0;
			} else if (units === "failures") {
				decimals = 0;
			} else if (units === "successes") {
				decimals = 0;
			} else if (units === "runs") {
				decimals = 0;
			} else if (units) {
				decimals = 1;
			}
			if (derivative === "PER_SECOND") {
				decimals += 1;
			}
			result += "{numDecimals:" + decimals + scale + "}";
			if (scale === "") {
				result += " " + units;
			}
			if (derivative === "PER_SECOND") {
				result += "/s";
			}
			return result;
		},

		/**
		* The start/end date is what we display by default.
		*/
		getExpandFactor: function() {
			return 1;
		},

		/**
		* We cache 3x of the visible data.
		*/
		getCacheFactor: function() {
			return 3;
		},

		initializeTime: function() {
			this._super.apply(this, arguments);
			// this needs to be done before the config has been initialized.
			if (this.allowChartsToStretch) {
				this.chartStretcher = new ChartStretcher();
			}
		},

		initializeData: function() {
			this._super.apply(this, arguments);
			// this needs to be done after the config has been initialized.
			if (this.chartStretcher) {
				this.chartStretcher.initialize(this);
			}
		},
		/**
		* Template method to allow sub classes to know that the chart's visible range has changed.
		* Note: this is different from _checkRange() defined in the parent class.
		*
		* _checkRange is called in two occasions:
		* * When the chart's visible range was changed directly by the user,
		* * When the chart's visible range was changed externally.
		*
		* _broadcastRangeChange is called in the first occasion:
		* * When the chart's visible range was changed directly by the user.
		*/
		_broadcastRangeChange: function() {
		},

		/**
		* Handles selection change.
		* We need to detect when it is a good time to broadcast time range changes.
		*
		* We should only do this when the event was initiated from this control only.
		*/
		onSelectedRangeChange: function(chartStartDate, chartEndDate, phase, sender) {
			this._super.apply(this, arguments);

			// The sender could be
			// mouseDrag (dragging),
			// scroller (moving the scroller),
			// external (externally calling zoomTo or scrollTo)

			// The phase could be "finish" or "progress".
			if (sender !== "external") {
				var broadcast = false;

				if (phase === "finish") {
					broadcast = true;
				} else if (phase === "progress") {
					if (sender === "mouseDrag" || sender === "scroller") {
						broadcast = true;
					}
				}
				if (broadcast) {
					this._broadcastRangeChange();
				}
			}
		},

		/**
		 * Renders the charts.
		 * Fetches data (asynchronous) if we need to fetch more data.
		 * Updates the chart settings.
		 */
		render: function() {
			if (this._timeSeriesAdded) {
				this._checkRange(true);
				this._timeSeriesAdded = false;
			}

			if (this._settingsChanged) {
				var result = this._applySettingsChanges();
				this._settingsChanged = false;
				return result;
			} else {
				return true;
			}
		},

		/**
		 * Updates the AnyChart's chart settings,
		 * which would redraw the charts.
		 */
		_applySettingsChanges: function() {
			// console.log("_applySettingsChanges");
			if (this.chart) {
				return this.chart.applySettingsChanges();
			}
			return true;
		},

		/**
		 * _checkRange checks if we need to load more data.
		 *
		 * From AnyChart's perspective, it has these four values in its data model:
		 *
		 * firstDate        is the first data point.
		 * firstVisibleDate is the first visible data point.
		 * lastVisibleDate  is the last visible data point.
		 * lastDate         is the last data point.
		 *
		 * From our perspective, we need 4 new variables:
		 * this._latestFirstFetchedDate and this._earliestLastFetchedDate.
		 * This range represents the range that we have "real" data from the server
		 * for all the time series.
		 *
		 * this._earliestFirstFetchedDate and this._latestLastFetchedDate.
		 * this range represents the total range that we could have data from
		 * any of the time series. Needed for removal.
		 *
		 * Invariant:
		 * this.firstDate (could contain fake zero data) <= this._earliestFirstFetchedDate <= this._latestFirstFetchedDate;
		 * this.lastDate  (could contain fake zero data >= this._earliestLastFetchedDate <= this._latestLastFetchedDate;
		 *
		 * @param forceUpdate - always fetch more data.
		 */
		_checkRange: function(forceUpdate, currentMode) {

			if (!this.fetching || forceUpdate) {

				var visibleTime = this.delta(this.firstVisibleDate, this.lastVisibleDate);
				var durationToFetch = visibleTime;
				var bufferBefore = this.firstVisibleDate.getTime() - this.firstDate.getTime();
				var bufferAfter = this.lastDate.getTime() - this.lastVisibleDate.getTime();
				var now = Util.getServerNow();
				var lastTillNow = this.delta(this.lastDate, now);
				var newFirstDate, newLastDate;
				var padLeft = false, padRight = false;

				if (forceUpdate) {
					// Fetching a new range.
					// Clear all the data points, unfortunately, AnyChart doesn't expose
					// an API for this.
					// For the new data, we want to pad both ends.
					// e.g. 0 0 0 newData newData newData 0 0 0.
					if (this._earliestFirstFetchedDate && this._latestLastFetchedDate) {
						this._cleanDataSetBeforeFetch(this._earliestFirstFetchedDate, this._latestLastFetchedDate);
					}
					padLeft = true;
					padRight = true;
					this.fetching = true;
					this._fetchMoreData(this.firstDate, this.lastDate, padLeft, padRight);

				} else if (bufferBefore < visibleTime / 4) {
					newFirstDate = new Date(this.firstDate.getTime() - durationToFetch);
					if (this._latestFirstFetchedDate) {
						newLastDate = this.min(now, this._latestFirstFetchedDate);
						this._cleanDataSetBeforeFetch(this.firstDate, new Date(this._latestFirstFetchedDate.getTime() - 1));
					} else {
						newLastDate = this.min(now, this.firstDate);
					}

					// we are fetching some data to the left of the current time range,
					// so we cannot pad the right hand of the newly fetched range.
					// e.g. 0 0 0 newData newData newData current current current ...
					padLeft = true;
					padRight = false;
					this.fetching = true;
					this._fetchMoreData(newFirstDate, newLastDate, padLeft, padRight);
					this._setCachedRange(new TimeRange(newFirstDate, this.lastDate));

				} else if (bufferAfter < visibleTime / 4 && (lastTillNow > 60000 || currentMode)) {
					// TODO: 60000 represents 1 minute.
					// this is a temporary workaround required to make sure
					// the time control doesn't fetch data in the future unnecessarily.

					if (this._earliestLastFetchedDate) {
						newFirstDate = this._earliestLastFetchedDate;
						this._cleanDataSetBeforeFetch(new Date(this._earliestLastFetchedDate.getTime() + 1), this.lastDate);
					} else {
						newFirstDate = this.lastDate;
					}
					newLastDate = this.min(now, new Date(this.lastDate.getTime() + durationToFetch));

					// we are fetching some data to the right of the current time range,
					// so we cannot pad the left hand of the newly fetched range.
					// e.g. current current newData newData newData 0 0 0
					padLeft = false;
					padRight = true;
					this.fetching = true;
					this._fetchMoreData(newFirstDate, newLastDate, padLeft, padRight);
					this._setCachedRange(new TimeRange(this.firstDate, newLastDate));
				}
			}
		},

		/**
		 * Removes all the data points between the specified start/end time (inclusive).
		 * @param startDate - the specified start date.
		 * @param endDate - the specified end time date.
		 */
		_cleanDataSetBeforeFetch: function(startDate, endDate) {

			// Assume the worst case scenario that we need to update all the time series.
			var timeSeriesDescriptorsToRemove = this.getAllEnabledTimeSeriesDescriptors();

			// we have crossed the number of fixed slots in the charting system.
			// AnyChart need to refresh everything, so we need to fetch all the
			// time series again.
			if (this.slotSizeChanged) {
				// setting this to empty means we need to refresh all the time series.
				this.newTimeSeriesDescriptors = [];
			}

			// If the new array is still non-empty, then we should only clean the selected
			// ones instead.
			if (this.newTimeSeriesDescriptors && this.newTimeSeriesDescriptors.length > 0) {
				timeSeriesDescriptorsToRemove = this.newTimeSeriesDescriptors;
			}

			var i, dataSetIds = [];
			for (i = 0; i < timeSeriesDescriptorsToRemove.length; i += 1) {
				var dataSetId = this.getDataSetId(timeSeriesDescriptorsToRemove[i]);
				dataSetIds.push(dataSetId);
			}
			this.removeDataSetsInRange(dataSetIds, new TimeRange(startDate, endDate));
		},

		removeAllDataInRange: function(range) {
			var i, allDataSets = this.getDataSets(), len = allDataSets.length, dataSetIds = [];
			for (i = 0; i < len; i += 1) {
				var dataSetId = allDataSets[i].id;
				dataSetIds.push(dataSetId);
			}
			this.removeDataSetsInRange(dataSetIds, range);
		},

		removeDataSetsInRange: function(dataSetIds, range) {
			var i, len = dataSetIds.length;
			for (i = 0; i < len; i += 1) {
				var dataSetId = dataSetIds[i];
				if (dataSetId !== null) {
					// The startDate, endDate are specified in our timezone.
					// Anychart operates in its own timezone.
					try {
						this.chart.removeDataRange(dataSetId, this.toChartDate(range.startDate), this.toChartDate(range.endDate));
					} catch (ex) {
						console.log(ex);
					}
				}
			}
		},

		/**
		 * Called when fetchMoreData fetches some data from the server.
		 */
		onFetchedMoreData: function() {
			this.zoomTo(this.firstVisibleDate, this.lastVisibleDate);
		},

		/**
		 * Called when fetchMoreData received an error from the server.
		 * Either the data is not JSON, or there was no data.
		 */
		onFetchError: function(data, startDate, endDate) {
			if (this.chartStretcher) {
				this.chartStretcher.setRange(this, new TimeRange(startDate, endDate));
				this.chart.commitDataChanges();
			}
		},

		/**
		 * Called when fetchMoreData received some CSV data from the server.
		 * This allows us to preprocess the CSV data before passing it
		 * to the graph API.
		 *
		 * Today, when we send a request for data between [time1, time2),
		 * the server returns a range smaller than it: [time1 + delta1, time2 - delta2)
		 * where 0 <= delta1 < (time2 - time1)
		 *  and  0 <= delta2 < (time2 - time1)
		 * This causes all kinds of strange behaviors.
		 *
		 * The solution is to pad the return data.
		 *
		 * For example, suppose we ask for data between [10, 20)
		 * We get back   13,1\n14,2\n18,3\n
		 * This is not good because for the time control, it forces the new selected range
		 * to be [13,18).
		 * Even worse, when we get back an empty set from the backend,
		 * This squashes the new selected range completely.
		 *
		 * Instead, we need to pad the data back to the original range.
		 * We cannot draw a straight line from (10,0) to (13,1)
		 * or from (18,3) to (20,0); instead, we need to plot a few more points around csvData:
		 * time1,0    time1+delta1-1,0   csvData,    time2-delta2+1,0,   time2,0
		 *
		 * However, this padding creates a second problem: gaps of artificially
		 * generated zeros.

		 * To remove the gaps, I introduced 4 new variables:
		 * this._latestFirstFetchedDate and this._earliestLastFetchedDate.
		 * This range represents the range that we have "real" data from the server
		 * for all the time series.
		 *
		 * this._earliestFirstFetchedDate and this._latestLastFetchedDate.
		 * this range represents the total range that we could have data from
		 * any of the time series. Needed for removal.
		 *
		 * Imagine we asked for data between time 5 and time 30.
		 * Imagine the server returned two time series:
		 * 10,3\n15,5\n20,6 and 11,2\n15,6\n25,4.
		 *
		 * In JSON notation, with t = time, v = value, we have:
		 * [{t:10, v:3}, {t:15, v:5}, {t:20, v:6}],
		 * [{t:11, v:2}, {t:15, v:6}, {t:25, v:4}].
		 *
		 * On the charts with padding, this becomes:
		 * [{t:5, v:0}, {t:9, v:0}, {t:10, v:3}, {t:15, v:5}, {t:20, v:6}], {t:21, v:0}, {t:30, v:0}]
		 * [{t:5, v:0}, {t:10, v:0}, {t:11, v:2}, {t:15, v:6}, {t:25, v:4}], {t:26, v:0}, {t:30, v:0}];
		 *
		 * This means we can only assume valid data between t:11 and t:20.
		 * Next time when the client need to get data say from t:1 to t:5,
		 * to prevent gaps from appearing in any of the graphs,
		 * we would need to fetch data from t:1, and t:11.
		 *
		 * Similarly, if we need to get data say from t:30 to t:40,
		 * we would need to use t:20 to t:40.
		 *
		 * Ideally, there is one valid range value for each time series, but this requires
		 * a lot more changes, This approach is a close approximation.
		 *
		 * One question you might ask, why not just fetch from t:30 to t:40,
		 * since we already know there is no data between t:20 to t:30 for time series 1
		 * and t:25 - t:30 for time series 2?
		 *
		 * Problem is, the time series data is generated with some lag from the current time.
		 * If the current time is say 30, the server may or may have data between
		 * t:20 - t:30 yet.
		 *
		 * @param csvData - the data in CSV format.
		 * @param startDate - the start date
		 * @param endDate - the end date.
		 * @param padLeft - if true, and the server's first data point
		 * is not the same as the startDate, pad the left hand side with 1 or 2 zero entries,
		 * one at startDate.getTime(), one at first data point's timestamp - 1.
		 * (1 entry when the two points are identical).
		 * We also need to recalculate this._latestFirstFetchedDate to keep the invariant.
		 *
		 * @param padRight - if true, and the server's last data point
		 * is not the same as the endDate, pad the right hand side with 0, 1 or 2 zero entries,
		 * one at last data point's timestamp + 1, one at endDate.getTime().
		 * (1 entry when the two points are identical).
		 * We also need to recalculate this._earliestLastFetchedDate to keep the invariant.
		 */
		preprocessData: function(csvData, startDate, endDate, padLeft, padRight) {
			var firstCommaPos = csvData.indexOf(",");
			if (firstCommaPos === -1) {
				// we didn't get any data point for this range.
				if (padLeft) {
					// we need to set the _latestFirstFetchedDate past the end date.
					// because we have 0, ... ,0, current current .
					this._latestFirstFetchedDate = new Date(endDate.getTime() + 1);
					this._earliestFirstFetchedDate = this._latestFirstFetchedDate;
				}

				if (padRight) {
					// we need to set the _earliestLastFetchedDate before the start date.
					// because we have current current 0, ... , 0.
					this._earliestLastFetchedDate = new Date(startDate.getTime() - 1);
					this._latestLastFetchedDate = this._earliestLastFetchedDate;
				}
				if (this.allowPadding) {
					return startDate.getTime() + ",0\n" + endDate.getTime() + ",0\n";
				} else {
					return "";
				}
			} else {
				var lastCommaPos = csvData.lastIndexOf(",");
				var lastTimePos = csvData.lastIndexOf("\n", lastCommaPos);

				var firstFetchedTime = parseInt(csvData.substring(0, firstCommaPos), 10);
				var lastFetchedTime = parseInt(csvData.substring(lastTimePos, lastCommaPos), 10);

				var result = [];

				if (padLeft) {
					// We need to calculate the new starting point.
					// To be conservative, we track the latest value.
					if (!this._latestFirstFetchedDate || this._latestFirstFetchedDate.getTime() < firstFetchedTime) {
						this._latestFirstFetchedDate = new Date(firstFetchedTime);
					}
					// We also track the earliest value so we can remove everything effectively.
					if (!this._earliestFirstFetchedDate || this._earliestFirstFetchedDate.getTime() > firstFetchedTime) {
						this._earliestFirstFetchedDate = new Date(firstFetchedTime);
					}
					if (this.allowPadding && startDate.getTime() !== firstFetchedTime) {
						if (startDate.getTime() !== firstFetchedTime-1) {
							result.push(startDate.getTime(), ",0\n");
						}
						result.push((firstFetchedTime-1), ",0\n");
					}
				}
				// Assumption: csvData should end with \n.
				result.push(csvData);
				if (padRight) {
					// We need to calculate the new ending point.
					// To be conservative, we track the earliest value.
					if (!this._earliestLastFetchedDate || this._earliestLastFetchedDate.getTime() > lastFetchedTime) {
						this._earliestLastFetchedDate = new Date(lastFetchedTime);
					}
					// We also track the latest value so we can remove everything effectively.
					if (!this._latestLastFetchedDate || this._latestLastFetchedDate.getTime() > lastFetchedTime) {
						this._latestLastFetchedDate = new Date(lastFetchedTime);
					}
					if (this.allowPadding && endDate.getTime() !== lastFetchedTime) {
						result.push((lastFetchedTime+1), ",0\n");
						if (startDate.getTime() !== firstFetchedTime-1) {
							result.push(endDate.getTime(), ",0\n");
						}
					}
				}
				return result.join("");
			}
		},

		mergeEventGroups: function(allResults, chartId, seriesId) {
			if (!allResults) {
				return [];
			}
			var i, j, len = allResults.length, existingGroups = this.getEventMarkers().groups;
			if (!existingGroups) {
				existingGroups = [];
			}
			for (i = 0; i < len; i += 1) {
				var result = allResults[i];
				var events = result.events;
				if (events && events.length > 0) {
					var eventsLen = events.length;
					for (j = 0; j < eventsLen; j += 1) {
						var evt = events[j];
						var evtDate = new Date(evt.date);
						evt.date = Util.humanizeTimeShort(evtDate);
						var event = {
								id: result.type + "_event_" + j,
								visible: true,
								date: this.toChartDate(evtDate),
								format: this.getEventMarkerFormat(result.type),
								attributes: evt,
								settings: {
									shape: {
										fill: {
											color: this.getEventColor(evt),
											type: "Solid"
										},
										enabled: true
									}
								}
						};
						events[j] = event;
					}
				}

				if (eventsLen > 0) {
					var existingGroup = this.getEntryById(existingGroups, result.type);
					var k, group;
					if (existingGroup) {
						existingGroup.chart = chartId;
						existingGroup.series = seriesId;
						existingGroup.events.concat(events);
					} else {
						group = {
								id: result.type,
								visible: true,
								chart: chartId,
								series: seriesId,
								position: this.getEventMarkerPosition(result.type),
								events: events
						};
						existingGroups.push(group);
					}
				}
			}
			return existingGroups;
		},

		createEventSettings: function() {
			return {
				allowSelect: true,
				tooltip: {
					enabled: true,
					format: this.getEventTooltipFormat()
				}
			};
		},

		getEventColor: function(event) {
			return Util.severity2Color(event.severity);
		},

		getEventTooltipFormat: function() {
			return "<textformat leading=\"4\"><b>Time: {%date}<br/>{%message}";
		},

		getEventMarkerPosition: function(groupType) {
			return "Axis";
		},

		getEventMarkerFormat: function(groupType) {
			return groupType.charAt(0);
		},

		createEventMarkersFromResults: function(allResults) {
			var charts = this.getCharts();
			if (charts.length === 0) {
				return;
			}
			var chartId = charts[0].id;

			var seriesList = this.getSeriesList(chartId);
			if (seriesList.length === 0) {
				return;
			}
			var seriesId = seriesList[0].id;

			var groups = this.mergeEventGroups(allResults, chartId, seriesId);

			var result = {
				'enabled': true,
				'groupingMode': 'PutInStack',
				'groups': groups,
				'settings': this.createEventSettings(),
				'stackSettings': {
					'mode': 'AlwaysExpanded'
				}
			};
			return result;
		},

		applyEventMarkersChanges: function(allResults) {
			var result = false;
			if (this.chart && this.chart.objectModel) {
				this.chart.objectModel.eventMarkers = this.createEventMarkersFromResults(allResults);
				result = this.chart.applyEventMarkersChanges();
			}
			return result;
		},

		/*
		 * Fetches all the time series using a single AJAX call.
		 *
		 * It performs a HTTP POST and gets back a JSON structure:
		 * [
		 *   {
		 *      "id": dataSetId,
		 *      "data": "someCSVData",
		 *   }, {
		 *     ...
		 *   }
		 * ]
		 *
		 * @param startDate - the start date.
		 * @param endDate - the end date.
		 * @param padLeft - if true, and the server's first data point
		 * is not the same as the startDate, pad the left hand side with 2 zero entries,
		 * one at startDate.getTime(), one at first data point's timestamp - 1.
		 *
		 * @param padRight - if true, and the server's last data point
		 * is not the same as the endDate, pad the right hand side with 2 zero entries,
		 * one at last data point's timestamp + 1, one at endDate.getTime()
		 */
		_fetchMoreData: function(startDate, endDate, padLeft, padRight) {
			// Still need to debug this.
			// console.log("_fetchMoreData from " + startDate);
			// console.log("_fetchMoreData   to " + endDate);

			// If we're testing, don't fire AJAX events to the server here.
			if (Util.getTestMode()) {
				return;
			}

			var self = this;

			if (this.slotSizeChanged) {
				// setting this to empty means we need to refresh all the time series.
				this.newTimeSeriesDescriptors = [];

				this.slotSizeChanged = false;
				// Refresh AnyChart.
				if (this.chart && this.chart.applyConfigChanges) {
					this.chart.applyConfigChanges();
				}
			}

			if (this.showEventMarkers) {
				jQuery.post(
					"/cmf/events/queryjson",
					self._generateEventsPostData(startDate, endDate),
					function(data) {
						var allResults = data.allResults;
						if (allResults) {
							self.applyEventMarkersChanges(allResults);
						}
					},
					'json'
				);
			}

			jQuery.post(
				self.options.jsonURL,
				// Revisit: For some reason, setting handleAs as "json" doesn't work.
				// handleAs: "json",
				self._generateChartsPostData(startDate, endDate),
				function(data) {
					try {
						// if the data is not JSON, this will throw an
						// exception.
						var i;
						var len = data.length;
						if (len === 0) {
							// If there is no data, this is also an
							// error case.
							self.onFetchError(data, startDate, endDate);
						} else {
							// we have a new range available.
							if (padLeft) {
								// We need to recalculate these two values.
								// erase them before the for loop.
								self._latestFirstFetchedDate = undefined;
								self._earliestFirstFetchedDate = undefined;
							}
							if (padRight) {
								// We need to recalculate these two values.
								// erase them before the for loop.
								self._latestLastFetcheDate = undefined;
								self._earliestLastFetchedDate = undefined;
							}
							if (self.chartStretcher) {
								self.chartStretcher.setRange(self, new TimeRange(startDate, endDate));
							}

							// Process each chart.
							for (i = 0; i < len; i += 1) {
								var dataSetId = self.getDataSetIdFromTimeSeriesDescriptorId(data[i].id);
								if (dataSetId !== null) {
									self.chart.appendData(dataSetId,
											self.preprocessData(data[i].data, startDate, endDate, padLeft, padRight));
								}
							}
							// tell AnyChart to update the charts now.
							self.chart.commitDataChanges();
							self.onFetchedMoreData();
						}
					} catch (ex) {
						// This happens when the firehose didn't
						// return any valid data.
						self.onFetchError(data, startDate, endDate);
					}
					self.fetching = false;
					return data;
				}, 'json').error(function() {
					self.onFetchError("", startDate, endDate);
				});
			// setting this to empty means we have processed these descriptors.
			this.newTimeSeriesDescriptors = [];
		},


		// These variables cannot be instantiated at the prototype level.
		// Instead, they need to be created at the instance level.
		// @see createData
		// _numOfSlotsToExpandEachTime: 10,

		/**
		 * Number of slots to use;
		 */
		// _numOfSlots: 0,

		/**
		 * Number of free slots remaining.
		 */
		// _numOfFreeSlots: 0,

		/**
		 * An temp array used to hold those timeSeriesDescriptors that need
		 * an immediate update.
		 */
		// newTimeSeriesDescriptors: null,

		/**
		 * An array of [timeSeriesDescriptors]. Some of the entries could be null.
		 * It is unlikely that this array will be large.
		 */
		// _usedSlots: null,

		/**
		 * An array of [boolean]. A value of true indicates that
		 * the corresponding entry is used.
		 *
		 * We need this array because apparently, we cannot simply remove
		 * a previously used timeSeriesDescriptor because AnyChart doesn't
		 * allow dataSet removals without refreshing the entire object model.
		 */
		// _usedMarkers: null,

		createData: function(){
			this.newTimeSeriesDescriptors = [];
			this._numOfSlotsToExpandEachTime = 10;
			this._usedSlots = [];
			this._usedMarkers = [];
			this._numOfSlots = 0;
			this._numOfFreeSlots = 0;
			this._initializeSlots(this._numOfSlotsToExpandEachTime);
			return this._super.apply(this, arguments);
		},

		/**
		 * Initialize more slots.
		 */
		_initializeSlots: function(size) {
			var i;
			for (i = 0; i < size; i+= 1) {
				this._usedSlots.push(null);
				this._usedMarkers.push(false);
			}
			this._numOfSlots += size;
			this._numOfFreeSlots += size;
		},

		/**
		 * @return the number o
		 */
		getNumberOfSlots: function() {
			return this._numOfSlots;
		},

		/**
		 * Expands the slot pool.
		 *
		 * Whenever the slot pool is full, we need to increase the pool size.
		 * We are increasing it linearly because the size of this pool is likely
		 * to be small (< 50).
		 *
		 * We are currently not compacting the pool because we cannot
		 * really free any timeSeriesDescriptor from the corresponding slot id,
		 * because AnyChart still have a reference to it. When I tried
		 * to remove the chart/series from AnyChart's data model, it always hanged.
		 */
		_expandSlots: function() {
			var i;
			var expandBy = this._numOfSlotsToExpandEachTime;
			var currentSize = this._numOfSlots;

			this._initializeSlots(expandBy);

			// Update AnyChart's data model.
			jQuery.merge(this.getGeneralDataProviders(),
					this._createGeneralDataProviders(expandBy, currentSize));
			jQuery.merge(this.getDataSets(),
					this._createDataSets(expandBy, currentSize));

			this.slotSizeChanged = true;
		},

		_createGeneralDataProviders: function(len, startIndex) {
			var i, result = [];
			for(i = 0; i < len; i += 1) {
				var dataProvider = this.createDataProvider("dp_" + (i + startIndex),
						"dataSet_" + (i + startIndex), 1);
				result.push(dataProvider);
			}
			return result;
		},

		/**
		 * Override to create a fixed number of data providers.
		 */
		createGeneralDataProviders: function() {
			return this._createGeneralDataProviders(this.getNumberOfSlots(), 0);
		},

		_createDataSets: function(len, startIndex) {
			var i, result = [];
			for(i = 0; i < len; i += 1) {
				var dataSet = this.createDataSet("dataSet_" + (i + startIndex));
				result.push(dataSet);
			}
			return result;
		},

		/**
		 * Override to create a fixed number of dataSets.
		 */
		createDataSets: function() {
			return this._createDataSets(this.getNumberOfSlots(), 0);
		},

		/**
		 * @return the ID of a timeSeriesDescriptor.
		 */
		getTimeSeriesDescriptorId: function(timeSeriesDescriptor) {
			return timeSeriesDescriptor.context.type + "_" +
				timeSeriesDescriptor.context.id + "_" +
				timeSeriesDescriptor.metric.id;
		},

		/**
		 * @return all the enabled timeSeriesDescriptors.
		 */
		getAllEnabledTimeSeriesDescriptors: function() {
			var i, len = this._usedSlots.length, result = [];
			for (i = 0; i < len; i += 1) {
				if (this._usedSlots[i] !== null && this._usedMarkers[i]) {
					result.push(this._usedSlots[i]);
				}
			}
			return result;
		},

		/**
		 * @return all the timeSeriesDescriptors (enabled or not) for a given
		 * chartId.
		 */
		getAllTimeSeriesDescriptorsOnChart: function(chartId) {
			var i, len = this._usedSlots.length, result = [];
			for (i = 0; i < len; i += 1) {
				if (this._usedSlots[i] !== null) {
					var tsd = this._usedSlots[i];
					if (tsd.chart.id === chartId) {
						result.push(this._usedSlots[i]);
					}
				}
			}
			return result;
		},

		/**
		 * @return an array of all the visible charts.
		 */
		getVisibleCharts: function() {
			var i, charts = this.getCharts(), len = charts.length, result = [];
			for (i = 0; i < len; i += 1) {
				var chart = charts[i];
				if (chart.enabled) {
					result.push(chart);
				}
			}
			return result;
		},

		/**
		 * @return the slot index of a timeSeriesDescriptor by its ID, or -1 if not found.
		 */
		_indexOfTimeSeriesDescriptorById: function(timeSeriesDescriptorId) {
			var i, len = this._usedSlots.length, result = -1;
			for (i = 0; i < len; i += 1) {
				if (this._usedSlots[i] !== null) {
					var slotId = this.getTimeSeriesDescriptorId(this._usedSlots[i]);
					if (slotId === timeSeriesDescriptorId) {
						result = i;
						break;
					}
				}
			}
			return result;
		},

		/**
		 * @return the slot index of a timeSeriesDescriptor, or -1 if not found.
		 */
		_indexOfTimeSeriesDescriptor: function(timeSeriesDescriptor) {
			var id = this.getTimeSeriesDescriptorId(timeSeriesDescriptor);
			return this._indexOfTimeSeriesDescriptorById(id);
		},

		/**
		 * Compares two timeSeriesDescriptors and returns true if they are equivalent.
		 * @return true if the two input parameters are equivalent.
		 */
		_compareTimeSeriesDescriptor: function(tsd1, tsd2) {
			if (tsd1 === null && tsd2 === null) {
				return true;
			} else if (tsd1 === null || tsd2 === null) {
				return false;
			}
			return this.getTimeSeriesDescriptorId(tsd1) ===
				this.getTimeSeriesDescriptorId(tsd2);
		},

		/**
		 * Adds a timeSeriesDescriptor into the slot.
		 * @return true if a new entry is added, false if the specified entry
		 * exists already.
		 */
		_addTimeSeriesDescriptor: function(timeSeriesDescriptor) {
			var i, len = this._usedSlots.length, result = false,
				index = this._indexOfTimeSeriesDescriptor(timeSeriesDescriptor);

			if (index === -1) {
				if (this._numOfFreeSlots === 0) {
					this._expandSlots();
					len = this._usedSlots.length;
				}

				for(i = 0; i < len; i += 1) {
					if (this._usedSlots[i] === null) {
						this._usedSlots[i] = timeSeriesDescriptor;
						this._numOfFreeSlots -= 1;
						this._usedMarkers[i] = true;
						result = true;
						break;
					}
				}
			} else {
				// It is possible that the time series descriptor has been added before.
				// Mark it as used again.
				this._usedMarkers[index] = true;
				result = true;
			}
			return result;
		},

		/**
		 * Toggles the enable/disable state of a timeSeriesDescriptor.
		 *
		 * @param timeSeriesDescriptor
		 * @param value - true to enable, false to disable.
		 * An enabled timeSeriesDescriptor will fetch data from the remote server.
		 * A disabled timeSeriesDescriptor will not.
		 */
		_enableDisableTimeSeriesDescriptor: function(timeSeriesDescriptor, value) {
			var result = false, index = this._indexOfTimeSeriesDescriptor(timeSeriesDescriptor);

			if (index !== -1) {
				this._usedMarkers[index] = value;
				result = true;
			}
			return result;
		},

		/**
		 * APIs for manipulating contexts and metrics.
		 */
		getDataProviderIdFromTimeSeriesDescriptorId: function(timeSeriesDescriptorId) {
			var index = this._indexOfTimeSeriesDescriptorById(timeSeriesDescriptorId);
			if (index !== -1) {
				return "dp_" + index;
			}
			return null;
		},

		getDataProviderId: function(timeSeriesDescriptor) {
			var id = this.getTimeSeriesDescriptorId(timeSeriesDescriptor);
			return this.getDataProviderIdFromTimeSeriesDescriptorId(id);
		},

		getDataSetIdFromTimeSeriesDescriptorId: function(timeSeriesDescriptorId) {
			var index = this._indexOfTimeSeriesDescriptorById(timeSeriesDescriptorId);
			if (index !== -1 && this._usedMarkers[index]) {
				return "dataSet_" + index;
			}
			return null;
		},

		getDataSetId: function(timeSeriesDescriptor) {
			var id = this.getTimeSeriesDescriptorId(timeSeriesDescriptor);
			return this.getDataSetIdFromTimeSeriesDescriptorId(id);
		},

		/**
		* Each chart could display multiple metrics.
		*/
		getChartId: function(chart) {
			return chart.id;
		},

		getChartType: function(chart) {
			return chart.discrete ? "StepLine" : "Line";
		},

		/**
		 * Gets the series id used in the chart.
		 */
		getSeriesId: function(timeSeriesDescriptor) {
			return "series_" + timeSeriesDescriptor.context.type + "_" + timeSeriesDescriptor.context.id + "_" + timeSeriesDescriptor.metric.id;
		},

		/**
		 * Adds a time series.
		 */
		addTimeSeries: function(timeSeriesDescriptor, position) {

			var context = timeSeriesDescriptor.context;
			var chart = timeSeriesDescriptor.chart;
			var metric = timeSeriesDescriptor.metric;
			var result = false;

			if (context.type === metric.context) {

				var timeSeriesDescriptorAdded = this._addTimeSeriesDescriptor(timeSeriesDescriptor);

				if (timeSeriesDescriptorAdded) {
					var dpId = this.getDataProviderId(timeSeriesDescriptor);

					// Create a chart if necessary.
					var chartId = timeSeriesDescriptor.chart.id;
					var chartTitle = timeSeriesDescriptor.chart.title;
					var charts = this.getCharts();
					var chartElem;
					var chartPos = this.getPositionById(charts, chartId);

					if (chartPos !== -1) {
						// TODO (andyao): If we are placing on an existing chart, then
						// extra axis may be required (when we place multiple time series
						// of different units onto the same chart.
						// However, this doesn't work yet.
						//
						chartElem = charts[chartPos];
						// this.addExtraAxisIfRequired(chartElem, metric.units, metrics.derivative);
					} else {
						// the chart needs to be created.
					        chartElem = this.createChart(chartId, chartTitle, metric.units, metric.derivative);
						chartPos = this.addUniqueEntry(charts, chartElem);
						this._settingsChanged = true;
					}

					if (position !== undefined) {
						this.moveEntry(this.getCharts(), chartPos, position);
						this._settingsChanged = true;
					}
					// console.log(this.getSettings());

					// Create a chart series.
					var chartType = this.getChartType(timeSeriesDescriptor.chart);
					var seriesId = this.getSeriesId(timeSeriesDescriptor);
					var seriesName = this.getSeriesName(timeSeriesDescriptor);
					var seriesList = this.getSeriesList(chartId);
					var color = this.getColor(timeSeriesDescriptor, seriesList.length);
					var seriesPos = this.addUniqueEntry(seriesList,
							this.createSeries(seriesId, dpId, seriesName, color, chartType, metric.units, metric.derivative));

					if (chartPos !== -1 || seriesPos !== -1) {
						// It is possible that the series has already been added before.
						// Ensure it is enabled.
						seriesList[seriesPos].enabled = true;
						this.newTimeSeriesDescriptors.push(timeSeriesDescriptor);
						this._timeSeriesAdded = true;
						this._settingsChanged = true;
						result = true;
					}
				}
			}
			return result;
		},

		/**
		 * Creates a time series.
		 * @param seriesId - identifies the time series.
		 * @param dataProviderId - identifies the data source.
		 * @param name - the name of the time series, which appears in the legend
		 * area (next to the chart title).
		 * @param color - the color used to render this time series.
		 * @param type - "StepLineArea" or "Area".
		 * @param units - the metric's units attribute.
		 * @param derivative - string of whether this is a derivative or not
		 */
	        createSeries: function(seriesId, dataProviderId, name, color, type, units, derivative) {
			var series = {
				'type': type,
				'color': color,
				'id': seriesId,
				'dataProvider': dataProviderId,
				'name': name,
				'lineSeries': {
					'thickness': 2,
					'color': color,
					'enabled': true
				},
				'tooltipSettings': {
					'enabled': true,
					'separatedTooltip': {
					        'format': name + ": " + this.formatUnits(units, derivative),
						'font': {
							'color': "#FFFFFF"
						},
						'background': {
							'enabled': true,
							'fill': {
								'color': color,
								'opacity': 0.8,
								'enabled': true
							}
						}
					}
				 },
				'legendItem': {
					'enabled': true,
					'labels': {
						'fallingColor': color,
						'risingColor': color,
						'format': name,
						'icon': {
							'type': 'Circle',
							'size': 10,
							'fill': {
								'color': color,
								'opacity': 0.8,
								'enabled': true,
								'type': 'Solid'
							}
						},
						'font': {
							'color': color
						},
						'missingPointsBehaviorControl': {
							'searchForExistingPoint': false,
							'hideElementOnMissingValue': false
						}
					}
				},
				'marker': {
					'enabled': true
				}
			};
			return series;
		},

		/**
		 * Creates an extra Axis.
		 * @param id - the axis id.
		 * @param formattedUnit - the format string used to
		 * format values on this axis.
		 */
		createExtraAxis: function(id, formattedUnit) {
			return {
				'id': id,
				'position': "Left",
				'labels': {
					'enabled': true,
					'position': "Inside",
					'format': formattedUnit
				}
			};
		},

		/**
		 * Adds extra axis if a chart has time series with different units.
		 * TODO (andyao): In reality, this doesn't work yet, and I don't know why.
		 */
	        addExtraAxisIfRequired: function(chart, units, derivative) {
		        var formattedUnit = this.formatUnits(units, derivative);
			if (chart &&
					chart.valueAxes &&
					chart.valueAxes.primary &&
					chart.valueAxes.primary.labels &&
					chart.valueAxes.primary.labels.format !== formattedUnit) {
				// the primary axis is not a match,
				// loop through all the existing extra axis and see if there is a match.
				var i, found = false;
				if (chart.valueAxes.extras) {
					var extras = chart.valueAxes.extras;
					for (i = 0; i < extras.length; i += 1) {
						var extraAxis = extras[i];
						if (extraAxis.labels.format === formattedUnit) {
							found = true;
							break;
						}
					}
				} else {
					// in case it is not defined.
					chart.valueAxes.extras = [];
				}

				// Add the extra axis.
				if (!found) {
					var extraAxesId = chart.id + "_extra" + chart.valueAxes.extras.length;
					var extraAxes = this.createExtraAxis(extraAxesId, formattedUnit);
					chart.valueAxes.extras.push(extraAxes);
				}
			}
		},

		getSeriesName: function(timeSeriesDescriptor) {
			return timeSeriesDescriptor.context.id + " " + timeSeriesDescriptor.metric.title;
		},

		/**
		 * Determines the color for a time series.
		 *
		 * This implementation chooses a color based on the index of the
		 * time series within a given chart. So the first time series
		 * is always blue, the second time series is always green, etc.
		 *
		 * @param timeSeriesDescriptor - describes the time series.
		 * @param index - the index of the time series within the chart.
		 * @return the color for this time series.
		 */
		getColor: function(timeSeriesDescriptor, index) {
			return Util.getChartColor(index);
		},

		/**
		 * Removes a time series.
		 */
		removeTimeSeries: function(timeSeriesDescriptor) {
			return this.hideTimeSeries(timeSeriesDescriptor);
			/**
			 * NEED TO REVISIT:
			 * Currently, we cannot call removeSeries or removeChart
			 * without crashing AnyChart somehow.
			 *
			var context = timeSeriesDescriptor.context;
			var chart = timeSeriesDescriptor.chart;
			var metric = timeSeriesDescriptor.metric;
			var result = false;

			if (context.type === metric.context) {
				var chartId = timeSeriesDescriptor.chart.id;

				var seriesId = this.getSeriesId(timeSeriesDescriptor);
				var seriesRemoved = this.removeSeries(chartId, seriesId);

				var chartRemoved = false;
				var seriesList = this.getSeriesList(chartId);
				if (seriesList !== null && seriesList.length === 0) {
					chartRemoved = this.removeChart(chartId);
				}

				this._disableTimeSeriesDescriptor(timeSeriesDescriptor);

				if (seriesRemoved || chartRemoved) {
					result = true;
				}
			}
			*/
		},

		_hideShowTimeSeries: function(timeSeriesDescriptor, value) {
			this._enableDisableTimeSeriesDescriptor(timeSeriesDescriptor, value);

			var seriesId = this.getSeriesId(timeSeriesDescriptor);
			var series = this.getSeries(timeSeriesDescriptor.chart.id, seriesId);
			if (series && series.enabled !== value) {
				series.enabled = value;
				if (value) {
					// when we show a timeSeries, we need to make sure it is fetched next.
					this.newTimeSeriesDescriptors.push(timeSeriesDescriptor);
					this._timeSeriesAdded = true;
				}
				this._settingsChanged = true;
				return true;
			}
			return false;
		},


		hideTimeSeries: function(timeSeriesDescriptor) {
			return this._hideShowTimeSeries(timeSeriesDescriptor, false);
		},

		showTimeSeries: function(timeSeriesDescriptor) {
			return this._hideShowTimeSeries(timeSeriesDescriptor, true);
		},

		/**
		 * Show/Hide a chart.
		 *
		 * Note: we do not disable all the timeSeries automatically.
		 */
		_hideShowChart: function(chartId, value) {
			var chart = this.findChartById(chartId);
			if (chart && chart.enabled !== value) {
				chart.enabled = value;
				this._settingsChanged = true;
				return true;
			}
			return false;
		},

		hideChart: function(chartId) {
			return this._hideShowChart(chartId, false);
		},

		/**
		 * Show a chart at a specified position.
		 */
		showChart: function(chartId, position) {
			var result = this._hideShowChart(chartId, true);
			if (position !== undefined) {
				var chart = this.getChart(chartId);
				if (chart !== null) {
					var chartPos = this.addUniqueEntry(this.getCharts(), chart);
					this.moveEntry(this.getCharts(), chartPos, position);
					this._settingsChanged = true;
					result = true;
				} else {
					result = false;
				}
			}
			return result;
		},

		/**
		 * @return true if we have 1 or more charts.
		 */
		hasCharts: function() {
			return this.getCharts().length > 0;
		},

		/**
		 * @return the chart object model if we have it.
		 */
		findChartById: function(chartId) {
			return this.getEntryById(this.getCharts(), chartId);
		},

		onChartDraw: function() {
			this._super.apply(this, arguments);
			if (!this.initialFetch) {
				var padLeft = true, padRight = true;
				this._fetchMoreData(this.firstDate, this.lastDate, padLeft, padRight);
				this.initialFetch = true;
			}
		},

		setShowEventMarkers : function(value) {
			this.showEventMarkers = value;
		}
	});

});
